#include "storage.hh"
#include "templates.hh"

#include <iostream>
#include <cstdlib>

#include <map>

namespace VectorStorage
{
//private:
    typedef unsigned short LengthType;
//public:
    class VecType
    {
    private:
        unsigned char* Buf;
        LengthType Length;
        bool owned, changed;
    private:
        /* No copying. */
        VecType(const VecType&);
        void operator=(const VecType&);
    public:
        VecType(): Buf(0),Length(0),owned(false),changed(false) { }
        ~VecType() { if(owned && Buf) std::free(Buf); }

        inline bool was_changed() const { return changed; }
        inline void reset_change() { changed=false; }
    
        inline const unsigned char operator[] (unsigned ind) const
        {
            return Buf[ind];
        }
        inline void initial_assign_readonly
            (const unsigned char* begin, const unsigned char* end)
        {
            if(owned && Buf) std::free(Buf);
            Buf    = const_cast<unsigned char*>(begin);
            Length = end-begin;
            owned  = false;
            changed= false;
        }
        inline void assign(const unsigned char* begin, const unsigned char* end)
        {
            unsigned NewLen = end-begin;
            if(!owned) { Buf = (unsigned char*)std::malloc(Length = NewLen); owned=true; }
            if(Length != NewLen)
            {
                if(Buf) std::free(Buf);
                Buf = (unsigned char*) std::malloc(Length = NewLen);
            }
            std::memcpy(Buf, begin, Length);
            changed=true;
        }
        inline bool empty() const { return Length == 0; }
        inline unsigned size() const { return Length; }
        inline void clear()
        {
            if(owned && Buf) std::free(Buf);
            Buf=0;
            Length=0;
            owned=false;
            changed=true;
        }
        void replace(unsigned beginpos, unsigned orig_size,
                     const unsigned char* source, unsigned new_size)
        {
            if(orig_size == Length)
            {
                /* The whole content is being replaced */
                assign(source, source+new_size);
                return;
            }
            if(!owned)
            {
                assign(Buf, Buf+Length);
            }
            
            unsigned tail_size = Length-(beginpos+orig_size);
            if(new_size > orig_size)
            {
                unsigned diff = new_size - orig_size;
                Buf = (unsigned char*)std::realloc(Buf, Length + diff);
                std::memmove(Buf+beginpos+new_size,
                             Buf+beginpos+orig_size,
                             tail_size);
                Length += diff;
            }
            else if(new_size < orig_size)
            {
                unsigned diff = orig_size - new_size;
                std::memmove(Buf+beginpos+new_size,
                             Buf+beginpos+orig_size,
                             tail_size);
                Buf = (unsigned char*)std::realloc(Buf, Length - diff);
                Length -= diff;
            }
            std::memcpy(Buf+beginpos, source, new_size); 
            changed=true;
        }
        inline const unsigned char* GetReadPtr(unsigned pos, unsigned length) const
        {
            if(!Length) return (const unsigned char*)"";
            return Buf+pos;
        }
        /*
        inline unsigned char* GetWritePtr(unsigned pos, unsigned length)
        {
            if(!Length) { static unsigned char c; return &c; }
            if(!owned)
            {
                assign(Buf, Buf+Length);
            }
            return Buf+pos;
        }
        */
    };

//public:
    /* The storage introduced by Read().
     * It acts like a vector, and it masquerades as the
     * actual storage of the SimpleVec item, but in fact
     * it's just a relay.
     */
    static VecType Buf;
//private:

    /* The storage where all data is stored in.
     * It is governed by the StupidVec items and clear().
     */
    static MemoryStorage<unsigned char> storage;

//private:
    class StupidVec
    {
        unsigned long Ptr;
        LengthType Length, AllocLength;
    public:
        StupidVec() : Ptr(0),Length(0),AllocLength(0) { }
        /* Warning: destructor is never called. */
        
        void Reassign(const VecType& vec)
        {
            if(unlikely(vec.size() > ((LengthType)-1)))
            {
                std::cerr << "Apparently, byte was not enough for StupidVec::Length!\n";
                abort();
            }
        
            
            if(vec.size() > AllocLength)
            {
                unsigned DesiredSize = (vec.size() + 3) & ~3;
#if 1
                static std::multimap<LengthType/*size*/,unsigned/*pos*/> holes;
                
                /*static unsigned counter=0;
                if(!counter--){counter=500000;
                 std::cerr << "holes.size()="<<holes.size() << "\n";
                }*/
                
                if(storage.empty()) holes.clear();
                if(AllocLength > 0)
                {
                    holes.insert(std::make_pair(AllocLength,Ptr));
                }
                std::multimap<LengthType,unsigned>::iterator h = holes.find(DesiredSize);
                if(h != holes.end())
                {
                    Ptr = h->second;
                    holes.erase(h);
                }
                else
#endif
                {
                    Ptr = storage.size();
                }
                AllocLength = DesiredSize;
            }
            Length = vec.size();
            
            if(likely(Length > 0))
            {
                unsigned char* target = storage.GetWritePtr(Ptr, AllocLength);
                std::memcpy(target, vec.GetReadPtr(0,Length), Length);
            }
        }
        const unsigned char* GetBuf() const
        {
            return storage.GetReadPtr(Ptr, Length);
        }
        unsigned GetLength() const { return Length; }
    } __attribute__((packed));
    
//public:
    typedef unsigned long IDtype;
//private:
    static IDtype BufOwner = 0;
    static IDtype ID = 0;
    static unsigned used = 0;
    
    /* The storage of pointers to StupidVec items.
     * Indexed by the "IDs" which are the public
     * interface to this allocator.
     * It is governed by Create(), Free1() and clear().
     */
    static MemoryStorage<StupidVec> ptrs;
//public:
    static IDtype Create()
    {
        ++used;
        
        StupidVec* ptr = ptrs.GetWritePtr(ID);
        // Construct the item in the buffer
        new(ptr) StupidVec;
        return ID++;
    }
    static void clear()
    {
        ID=0;
        ptrs.clear();
        storage.clear();
        BufOwner=0; Buf.clear();
        used=0;
    }
    static void Free1()
    {
        --used;
        if(!used) clear();
    }
    
    /* References returned by this call are guaranteed to be
     * valid only until the next call of Read(), Create() or Free1().
     */
    static VecType& Read(IDtype n)
    {
        if(n != BufOwner)
        {
            if(Buf.was_changed())
            {
                ptrs.GetWritePtr(BufOwner)->Reassign(Buf);
            }
            const StupidVec& v = *ptrs.GetReadPtr(BufOwner=n);
            
            const unsigned char* Source = v.GetBuf();
            Buf.initial_assign_readonly(Source, Source+v.GetLength());
        }
        return Buf;
    }
}/*;

unsigned VectorStorage::used = 0;
MemoryStorage<unsigned char> VectorStorage::storage;
MemoryStorage<VectorStorage::StupidVec> VectorStorage::ptrs;
VectorStorage::VecType VectorStorage::Buf;
VectorStorage::IDtype VectorStorage::BufOwner = 0;
VectorStorage::IDtype VectorStorage::ID = 0;
*/
